use clap::{Parser, Subcommand};
use ed25519_dalek::{ed25519::signature::SignerMut, Signature, SigningKey, VerifyingKey};
use minicbor;
use pallas_primitives::PlutusData;
use rand::prelude::*;
use rand_chacha::rand_core::{RngCore, SeedableRng};
use serde::{Deserialize, Serialize};

#[derive(Parser)]
#[command(arg_required_else_help(true), version, about)]
struct Args {
    #[command(subcommand)]
    cmd: Option<Command>,
}

#[derive(Subcommand)]
enum Command {
    /// Gen a keypair from seed some sensible ephemeral key args to be used with `add`
    GenKeys { seed: Option<u64> },
    /// Sign an iou
    Sign {
        /// Signing key (hex encoded)
        skey: String,
        /// Tag (hex encoded)
        tag: String,
        /// Amount (unsigned integer u64. Seems to be largest permitted)
        amt: u64,
    },
    /// Verify an iou signature
    Verify {
        /// Verifying key (hex encoded)
        vkey: String,
        /// Tag (hex encoded)
        tag: String,
        /// Amount (unsigned integer u64. Seems to be largest permitted)
        amt: u64,
        /// Signature (hex encoded)
        sig: String,
    },
    /// Generate a random signed iou to test against aiken
    GenSignedIou { seed: Option<u64> },
}

fn main() -> anyhow::Result<()> {
    let args = Args::parse();

    match args.cmd {
        Some(Command::GenKeys { seed }) => {
            let skey = skey_from_seed(seed)?;
            println!("{}", serde_json::to_string(&KeyPairHex::from_skey(&skey))?);
        }
        Some(Command::Sign { skey, tag, amt }) => {
            let skey = skey_from_hex(&skey)?;
            let iou = Iou::from_cli(tag, amt)?;
            let msg = iou.to_vec()?;
            let sig = sign(&mut skey.clone(), msg.clone());
            let _ver = verify(&skey.verifying_key(), &msg, &sig);
            println!("{}", sig_to_hex(&sig));
        }
        Some(Command::Verify {
            vkey,
            tag,
            amt,
            sig,
        }) => {
            let vkey = vkey_from_hex(&vkey)?;
            let iou = Iou::from_cli(tag, amt)?;
            let msg = iou.to_vec()?;
            let sig = sig_from_hex(&sig)?;
            let res = verify(&vkey, &msg, &sig)?;
            println!("{:?}", res);
        }
        Some(Command::GenSignedIou { seed }) => {
            let siou = SignedIou::gen(seed)?;
            // println!("{}", serde_json::to_string(&siou)?) ;
            //println!("{}", serde_json::to_string(&SignedIou::gen(seed )?)?) ;
            println!("");
            siou.aiken_fmt(seed);
            println!("");
            aiken_test(seed);
            println!("");
        }
        None => {
            println!("See help");
        }
    }

    Ok(())
}

pub fn arr32(bytes: Vec<u8>) -> [u8; 32] {
    let arr: [u8; 32] = TryInto::try_into(bytes).expect("");
    arr
}

pub fn arr64(bytes: Vec<u8>) -> [u8; 64] {
    let arr: [u8; 64] = TryInto::try_into(bytes).expect("");
    arr
}

pub fn vkey_from_bytes(bytes: Vec<u8>) -> Result<VerifyingKey, anyhow::Error> {
    Ok(VerifyingKey::from_bytes(&arr32(bytes))?)
}

pub fn vkey_to_bytes(key: &VerifyingKey) -> Vec<u8> {
    Into::<Vec<u8>>::into(key.to_bytes())
}

pub fn vkey_from_hex(s: &str) -> Result<VerifyingKey, anyhow::Error> {
    vkey_from_bytes(hex::decode(s)?)
}

pub fn vkey_to_hex(key: &VerifyingKey) -> String {
    hex::encode(key.to_bytes())
}

pub fn seed_arr(seed: Option<u64>) -> [u8; 32] {
    let seed = match seed {
        Some(seed) => seed,
        None => rand::random::<u64>(),
    };
    let mut as_bytes: Vec<u8> = seed.to_le_bytes().into();
    as_bytes.resize(32, 0);
    arr32(as_bytes)
}

pub fn skey_from_seed(seed: Option<u64>) -> Result<SigningKey, anyhow::Error> {
    let arr = seed_arr(seed);
    Ok(SigningKey::from_bytes(&arr))
}

pub fn skey_from_bytes(bytes: Vec<u8>) -> Result<SigningKey, anyhow::Error> {
    Ok(SigningKey::from_bytes(&arr32(bytes)))
}

pub fn skey_from_hex(s: &str) -> Result<SigningKey, anyhow::Error> {
    skey_from_bytes(hex::decode(s)?)
}

pub fn skey_to_hex(key: &SigningKey) -> String {
    hex::encode(key.to_bytes())
}

pub fn sign(key: &mut SigningKey, msg: Vec<u8>) -> Signature {
    key.try_sign(&msg).unwrap()
}

pub fn verify(key: &VerifyingKey, msg: &Vec<u8>, sig: &Signature) -> Result<(), anyhow::Error> {
    let res = key.verify_strict(msg, sig)?;
    Ok(res)
}

pub fn sig_from_bytes(bytes: Vec<u8>) -> Result<Signature, anyhow::Error> {
    Ok(Signature::try_from(&arr64(bytes))?)
}

pub fn sig_to_bytes(sig: &Signature) -> Vec<u8> {
    Into::<Vec<u8>>::into(sig.to_bytes())
}

pub fn sig_from_hex(s: &str) -> Result<Signature, anyhow::Error> {
    sig_from_bytes(hex::decode(s)?)
}

pub fn sig_to_hex(sig: &Signature) -> String {
    hex::encode(sig.to_bytes())
}

#[derive(Serialize, Deserialize)]
struct KeyPairHex {
    skey: String,
    vkey: String,
}

impl KeyPairHex {
    fn from_skey(skey: &SigningKey) -> Self {
        Self {
            skey: skey_to_hex(&skey.clone()),
            vkey: vkey_to_hex(&skey.verifying_key()),
        }
    }
}

pub fn tag_from_hex(s: &str) -> Result<Vec<u8>, anyhow::Error> {
    let tag = hex::decode(s)?;
    Ok(tag)
}

#[derive(Serialize, Deserialize)]
struct Iou {
    tag: Vec<u8>,
    amount: u64,
}

impl Iou {
    fn from_cli(tag: String, amt: u64) -> Result<Self, anyhow::Error> {
        Ok(Self {
            tag: tag_from_hex(&tag)?,
            amount: amt,
        })
    }

    fn to_plutus_data(self: &Self) -> Result<PlutusData, anyhow::Error> {
        let data = PlutusData::Constr(pallas_primitives::Constr {
            tag: 121,
            any_constructor: None,
            fields: pallas_primitives::MaybeIndefArray::Indef(vec![
                PlutusData::BoundedBytes(pallas_primitives::BoundedBytes::from(self.tag.clone())),
                PlutusData::BigInt(pallas_primitives::BigInt::Int(
                    pallas_primitives::Int::try_from(self.amount as i128)?,
                )),
            ]),
        });
        Ok(data)
    }

    fn to_vec(self: &Self) -> Result<Vec<u8>, anyhow::Error> {
        let v = minicbor::to_vec(self.to_plutus_data()?)?;
        Ok(v)
    }

    fn gen(seed: Option<u64>) -> Self {
        let mut rng: rand_chacha::ChaCha8Rng = SeedableRng::seed_from_u64(seed.unwrap_or(0));
        let tag: Vec<u8> = rng.next_u64().to_le_bytes().into();
        let amount = rng.next_u64();
        Iou { tag, amount }
    }
}

#[derive(Serialize, Deserialize)]
struct SignedIou {
    vkey: String,
    tag: String,
    amount: u64,
    msg: String,
    sig: String,
}

impl SignedIou {
    fn gen(seed: Option<u64>) -> Result<Self, anyhow::Error> {
        let skey = skey_from_seed(seed)?;
        let vkey = skey.verifying_key();
        let iou = Iou::gen(seed);
        let msg = iou.to_vec()?;
        let sig = sign(&mut skey.clone(), msg.clone());
        Ok(Self {
            vkey: vkey_to_hex(&vkey),
            tag: hex::encode(iou.tag),
            amount: iou.amount,
            msg: hex::encode(&msg),
            sig: sig_to_hex(&sig),
        })
    }

    fn aiken_fmt(self: Self, seed: Option<u64>) {
        let Self {
            vkey,
            tag,
            amount,
            msg,
            sig,
        } = self;
        println!("const i{} = IouTest {{", seed.unwrap_or(0),);
        println!("  vkey: {}", aiken_hex(&vkey));
        println!("  tag: {}", aiken_hex(&tag));
        println!("  amount: {},", amount);
        println!("  msg: {}", aiken_hex(&msg));
        println!("  sig: {}", aiken_hex(&sig));
        println!("}}");
    }
}

fn aiken_test(seed: Option<u64>) {
    println!("test test_i{} () {{", seed.unwrap_or(0));
    println!("  let i = i{}", seed.unwrap_or(0));
    println!("  verify_iou(i.vkey, i.tag, t.Iou (i.amount, i.sig))");
    println!("}}");
}

fn aiken_hex<T>(thing: &T) -> String
where
    T: ToString,
{
    format!("#\"{}\"", thing.to_string())
}

// Want scenario.
// This is (step, input, output)
